/**
 * P4
 * http://snarfed.org/space/p4
 *
 * Copyright 2001-2004
 * Ken Ashcraft <kashcraft+p4@gmail.com>
 * Ryan Barrett <p4@ryanb.org>
 * Maulik Shah <p4@maulik.net>
 * Nathan Stoll <nathanstoll+p4@gmail.com>
 *
 * @file msg_hello.cpp
 *
 * Defines the cmsg_hello class, which is derived from the abstract cmsg class.
 * The cmsg_hello class is the class that HELLO messages instantiate. For more
 * information, see protocol.txt, msg.h, and protocol.h.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include "conn_mgr.h"
#include "msg_hello.h"
#include "receiveq.h"
#include "lka.h"
#include "id_table.h"
#include "p4error.h"
#include "debug.h"

#ifdef _WIN32
#include <winsock.h>
#else
#include <netinet/in.h>  // for struct sockaddr_in
#endif

/*****************************************************************************
 * PUBLIC METHODS
 *****************************************************************************/
cmsg_hello::cmsg_hello() : cmsg(0, 0, 0)
{
  hdr.set_type(kp4_hello); 
  hdr.set_flag(kp4_broadcast); 

  creceiveq &rq = creceiveq::instance();

  // allocate data
  ushort num_plugins = rq.num_plugins();
  hdr.data_len = sizeof(p4_hello) + num_plugins * sizeof(plugin_info);
  host_data_len = hdr.data_len;
  data = new char[hdr.data_len];

  hello = reinterpret_cast<p4_hello *>(data);
  plugins = reinterpret_cast<plugin_info *>(data + sizeof(p4_hello));


  // fill in address
  struct sockaddr_in sin;
  if (get_local_addr(sin) != kerr_success)
    cdebug::instance() << acquire("msg_debug")
                       << "COULDN'T GET THE LOCAL ADDRESS! "
                       << "ARE YOU TOO GOOD FOR YOUR HOME?!?"
                       << release();

  hello->addr = sin.sin_addr.s_addr;
  hello->port = csettings::instance().get<ushort>("p4_port");


  // fill in plugin ids
  ushort num_written;
  uint ret = rq.get_plugins(num_plugins, plugins, num_written);
  assert(ret == kerr_success && num_plugins == num_written);
  hello->num_plugins = num_plugins;

  debug_print();
}

// this only gets called by servicer. we need to handle the message coming back
// to us.
const cconn &cmsg_hello::send(const string &username)
{
  cid_table::instance().add(hdr.message_id);
  return cmsg::send(username); 
}

/*****************************************************************************
 * PROTECTED METHODS
 *****************************************************************************/
cmsg_hello::cmsg_hello(const p4_header &hdr) : cmsg(hdr)
{
  hello = reinterpret_cast<p4_hello *>(data);
  plugins = reinterpret_cast<plugin_info *>(data + sizeof(p4_hello));
}

cmsg_hello::cmsg_hello(const cmsg &msg) : cmsg(msg)
{
  hello = reinterpret_cast<p4_hello *>(data);
  plugins = reinterpret_cast<plugin_info *>(data + sizeof(p4_hello));
}

bool cmsg_hello::virt_process() const
{
  debug_print();

  clka::instance().add(hdr.username, hello->addr, hello->port,
                       hello->num_plugins, plugins);
  return true;
}

void cmsg_hello::debug_print() const
{
  assert(data);

  cdebug::instance() << acquire("msg_debug")
                     << "Hello from " << hdr.username << " with "
                     << hello->num_plugins
                     << " plugins:";

  for (int i = 0; i < hello->num_plugins; i++)
    cdebug::instance() << " " << plugins[i].id;

  cdebug::instance() << release();
}


/*****************************************************************************
 * PRIVATE METHODS
 *****************************************************************************/
void cmsg_hello::convert_byteorder(cmsg::short_fn sfn, cmsg::long_fn lfn)
{
  cmsg::convert_byteorder(sfn, lfn);

  assert(hello && plugins);

  hello->addr = lfn(hello->addr);
  hello->port = sfn(hello->port);
  hello->num_plugins = sfn(hello->num_plugins);

  // get num_plugins in host order (this looks backwards because this is called
  // from cmsg::hton or ntoh, and byte_order hasn't been flipped yet)
  ushort num_plugins = ((byte_order == ehost) ?
                        ntohs(hello->num_plugins) : hello->num_plugins);
  for (int i = 0; i < num_plugins; i++)
    plugins[i].id = sfn(plugins[i].id);
}
